#define INCL_DOS
#define INCL_WIN
#define INCL_DOSERRORS
#define INCL_MCIOS2
#define INCL_MACHDR
#define INCL_MMIOOS2

#include <os2.h>
#include <os2me.h>
#include <string.h>
#include <process.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>

#include "oscardata.h"
#include "oscarsession.h"
#include "compatibility.h"

#ifdef MRM_DEBUG_PRINTF
int debugf( char *fmt, ... )
{
  if ( debugOutputStream )
  {
    int retVal;
    va_list argptr;
    
    fprintf( debugOutputStream, "[TID%02d]: ", (*_threadid) );
    va_start( argptr, fmt );
    retVal = vfprintf( debugOutputStream, fmt, argptr );
    va_end( argptr );
    fflush( debugOutputStream );
    
    return retVal;
  }
  return 0;
}
#endif

#ifdef MRM_DEBUG_MEMORY
blockHistory *memTracer = NULL;
int numMemTracers = 0;
int allocMemTracers = 0;
#endif

static ULONG (EXPENTRY *pMciSendCommand) ( USHORT usDeviceID,
 USHORT usMessage, ULONG ulParam1, PVOID pParam2, USHORT usUserParm ) = NULL;

static HMMIO (EXPENTRY *pMmioOpen) ( const char *pszFileName,
 PMMIOINFO pmmioinfo, ULONG ulOpenFlags ) = NULL;
static USHORT (EXPENTRY *pMmioClose) ( HMMIO hmmio, USHORT usFlags ) = NULL;
static ULONG (EXPENTRY *pMmioIdentifyFile) ( PSZ pszFileName,
 PMMIOINFO pmmioinfo, PMMFORMATINFO pmmformatinfo, PFOURCC pfccStorageSystem,
 ULONG ulReserved, ULONG ulFlags ) = NULL;

static HMODULE mmpmDLL = 0;
static HMODULE mdmDLL = 0;

static ULONG numWaveOuts = 0;
static ULONG numMIDIouts = 0;
static ULONG *mmThreads = NULL;
static USHORT *mmDeviceIDs = NULL;
static ULONG numMmThreads = 0;
static ULONG mmThreadsSize = 0;
static HEV mmMutex = 0;
static ULONG audioVolume = 100;

char **waveOutNames, **midiOutNames;

#ifdef MRM_DEBUG_MEMORY

void CHECKED_FREE_TRACER( void *ptr, const char *file, int line, int flags )
{
  int i;
  
  if ( !ptr )
  {
    debugf( "MEMORY MANAGEMENT FAULT!!: FREE NULL POINTER\n" );
    debugf( "ATTEMPTED TO FREE NULL POINTER IN %s LINE %d.\n",
     file, line );
    return;
  }
  
  for ( i=0; i<numMemTracers; ++i )
  {
    if ( !memTracer[i].blockAddr ) continue;
    
    if ( memTracer[i].blockAddr == ptr )
    {
      if ( memTracer[i].freeModule )
      {
        debugf( "MEMORY MANAGEMENT FAULT!!: MULTIPLE FREES (0x%08lx)\n",
         ptr );
        debugf( "BLOCK ALLOCATED IN %s LINE %d.\n", memTracer[i].allocModule,
         memTracer[i].allocLine );
        debugf( "BLOCK DEALLOCATED IN %s LINE %d.\n", memTracer[i].freeModule,
         memTracer[i].freeLine );
        debugf( "ADDITIONAL FREE IN %s LINE %d.\n", file, line );
        return;
      } else {
        if ( !(flags & MRM_DEBUG_MEMCHECK_ISREALLOC) )
          free( ptr );
        
        memTracer[i].freeModule = file;
        memTracer[i].freeLine = line;
        DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT,
         &memTracer[i].timestamp, 4 );
        return;
      }
    }
  }
  
  debugf( "MEMORY MANAGEMENT FAULT!!: FREE A NON-ALLOCATED BLOCK (0x%08lx)\n",
   ptr );
  debugf( "BAD FREE IN %s LINE %d.\n", file, line );
}

void *CHECKED_MALLOC_TRACER( int size, const char *file, int line, int flags,
 void *reallocPtr )
{
  void *retPtr;
  int i, firstFree = numMemTracers, oldest = numMemTracers;
  
  if ( !size )
  {
    debugf( "MEMORY MANAGEMENT FAULT!!: ALLOC OF SIZE 0\n" );
    debugf( "ALLOC IN %s LINE %d.\n", file, line );
    return NULL;
  }
  
  if ( flags & MRM_DEBUG_MEMCHECK_ISCALLOC )
  {
    retPtr = calloc( size, 1 );
  } else {
    if ( flags & MRM_DEBUG_MEMCHECK_ISREALLOC )
    {
      retPtr = realloc( reallocPtr, size );
    } else {
      retPtr = malloc( size );
    }
  }
  
  for ( i=0; i<numMemTracers; ++i )
  {
    if ( !memTracer[i].blockAddr ) continue;
    
    if ( memTracer[i].blockAddr == retPtr )
    {
      memTracer[i].allocModule = file;
      memTracer[i].allocLine = line;
      memTracer[i].freeModule = NULL;
      memTracer[i].freeLine = 0;
      DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT,
       &memTracer[i].timestamp, 4 );
      return retPtr;
    } else if ( firstFree == numMemTracers && !memTracer[i].blockAddr )
    {
      firstFree = i;
    }
    if ( (oldest == numMemTracers ||
          memTracer[i].timestamp < memTracer[oldest].timestamp) &&
          memTracer[i].allocModule && memTracer[i].freeModule )
    {
      oldest = i;
    }
  }
  
  if ( firstFree == numMemTracers )
  {
    if ( numMemTracers == allocMemTracers )
    {
      // Take the least recently accessed block that has been cleaned up
      //  and remove it from our tracer list.  This could cause some double
      //  frees to be seen as bad frees, but it will keep our tracer size
      //  from growing to ludicrous proportions.
      
      if ( oldest == numMemTracers )
      {
        debugf( "WARNING: Memory tracer history is full.  This may produce false hits\n" );
        debugf( "         pertaining to the block allocated at 0x%08lx.\n",
         retPtr );
        return retPtr;
      } else {
        firstFree = oldest;
      }
    } else {
      numMemTracers++;
    }
  }
  
  memTracer[firstFree].blockAddr = retPtr;
  memTracer[firstFree].allocModule = file;
  memTracer[firstFree].allocLine = line;
  memTracer[firstFree].freeModule = NULL;
  memTracer[firstFree].freeLine = 0;
  DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT,
   &memTracer[firstFree].timestamp, 4 );
  
  return retPtr;
}

#endif

static void soundErrorCleanup( void )
{
  if ( mmpmDLL )
    DosFreeModule( mmpmDLL );
  if ( mdmDLL )
    DosFreeModule( mdmDLL );
  
  mmpmDLL = 0;
  mdmDLL = 0;
  pMciSendCommand = NULL;
  pMmioOpen = NULL;
  pMmioClose = NULL;
}

char *getDeviceDescription( USHORT devType, USHORT devOrdinal )
{
  MCI_SYSINFO_PARMS siParms = { 0 };
  MCI_OPEN_PARMS openParms = { 0 };
  MCI_INFO_PARMS infoParms = { 0 };
  MCI_GENERIC_PARMS genParms = { 0 };
  ULONG rc, bufSize;
  char *outbuf;
  
  CHECKED_CALLOC( 128, outbuf );
  bufSize = 128;
  
  openParms.pszDeviceType = (PSZ) (MAKEULONG( devType, devOrdinal ));
  
  rc = (*pMciSendCommand) ( 0, MCI_OPEN,
   MCI_OPEN_TYPE_ID | MCI_OPEN_SHAREABLE | MCI_WAIT, &openParms, 0 );
  
  if ( rc )
  {
    if ( rc == MCIERR_DEVICE_LOCKED )
    {
      debugf( "Device type %u, ordinal %u is locked by another process.\n",
       devType, devOrdinal );
    } else {
      debugf( "MCI_OPEN failed for device type %u, ordinal %u.  RC = %lu\n",
       devType, devOrdinal, rc );
    }
    CHECKED_FREE( outbuf );
    return NULL;
  }
  
  infoParms.pszReturn = outbuf;
  infoParms.ulRetSize = bufSize;
  
  rc = (*pMciSendCommand) ( openParms.usDeviceID, MCI_INFO,
   MCI_INFO_PRODUCT | MCI_WAIT, &infoParms, 0 );
  
  if ( rc == MCIERR_INVALID_BUFFER )
  {
    CHECKED_REALLOC( outbuf, infoParms.ulRetSize, outbuf );
    bufSize = infoParms.ulRetSize;
    infoParms.pszReturn = outbuf;
    infoParms.ulRetSize = bufSize;
    
    rc = (*pMciSendCommand) ( openParms.usDeviceID, MCI_INFO,
     MCI_INFO_PRODUCT | MCI_WAIT, &infoParms, 0 );
  }
  
  rc = (*pMciSendCommand) ( openParms.usDeviceID, MCI_CLOSE,
   MCI_WAIT, &genParms, 0 );
  
  if ( outbuf && *outbuf )
  {
    return outbuf;
  }
  
  siParms.usDeviceType = devType;
  siParms.ulNumber = devOrdinal;
  siParms.pszReturn = outbuf;
  siParms.ulRetSize = bufSize;
  
  rc = (*pMciSendCommand) ( 0, MCI_SYSINFO,
   MCI_SYSINFO_INSTALLNAME | MCI_WAIT, &siParms, 0 );
  
  if ( rc == MCIERR_INVALID_BUFFER )
  {
    CHECKED_REALLOC( outbuf, siParms.ulRetSize, outbuf );
    bufSize = siParms.ulRetSize;
    siParms.pszReturn = outbuf;
    siParms.ulRetSize = bufSize;
    
    rc = (*pMciSendCommand) ( 0, MCI_SYSINFO,
     MCI_SYSINFO_INSTALLNAME | MCI_WAIT, &siParms, 0 );
  }
  
  if ( outbuf && *outbuf )
  {
    return outbuf;
  }
  
  debugf( "The Media Control Interface could not give me a name for device type %u, ordinal %u (rc=%lu).\n",
   devType, devOrdinal, rc );
  
  CHECKED_FREE( outbuf );
  
  return NULL;
}

void initializeCompatibility( int initMMPM )
{
  ULONG rc, i;

  #ifdef MRM_DEBUG_MEMORY
  
  memTracer = (blockHistory *) malloc( sizeof( blockHistory ) * 1024 );
  numMemTracers = 0;
  allocMemTracers = MRM_DEBUG_MAX_TRACERS;
  
  #endif
  
  if ( !initMMPM )
  {
    debugf( "Sound was disabled by a user command line option.\n" );
  } else {
    rc = DosLoadModule( NULL, 0, "MMPM.DLL", &mmpmDLL );
    
    if ( rc == 0 )
      rc = DosLoadModule( NULL, 0, "MDM.DLL", &mdmDLL );
    
    if ( rc == 0 )
    {
      rc = DosQueryProcAddr( mdmDLL, 1, NULL, (PFN *) &pMciSendCommand );
      rc |= DosQueryProcAddr( mmpmDLL, 40, NULL, (PFN *) &pMmioOpen );
      rc |= DosQueryProcAddr( mmpmDLL, 45, NULL, (PFN *) &pMmioClose );
      rc |= DosQueryProcAddr( mmpmDLL, 92, NULL, (PFN *) &pMmioIdentifyFile );
    }
    
    if ( rc )
    {
      debugf( "Unable to load MultiMedia Presentation Manager (MMPM.DLL).\n" );
      debugf( "Sound will be disabled as a result.\n" );
      
      soundErrorCleanup();
      return;
    } else {
      MCI_SYSINFO_PARMS mmInfo = { 0 };
      char outbuf[16];
      
      debugf( "MultiMedia Presentation Manager was initialized at runtime.\n" );
      
      mmInfo.usDeviceType = MCI_DEVTYPE_WAVEFORM_AUDIO;
      mmInfo.pszReturn = outbuf;
      mmInfo.ulRetSize = 16;
      
      rc = (*pMciSendCommand) ( 0, MCI_SYSINFO,
       MCI_SYSINFO_QUANTITY | MCI_WAIT, &mmInfo, 0 );
      
      if ( rc )
      {
        CHECKED_FREE( outbuf );
        debugf( "Error querying number of waveform audio devices (rc=%lu).\n",
         rc );
        debugf( "Waveform audio output is disabled.\n" );
        numWaveOuts = 0;
        waveOutNames = NULL;
      } else {
        numWaveOuts = atol( outbuf );
        if ( numWaveOuts )
        {
          CHECKED_CALLOC( sizeof( char * ) * numWaveOuts, waveOutNames );
        } else waveOutNames = NULL;
      }
      
      mmInfo.usDeviceType = MCI_DEVTYPE_SEQUENCER;
      mmInfo.pszReturn = outbuf;
      mmInfo.ulRetSize = 16;
      
      rc = (*pMciSendCommand) ( 0, MCI_SYSINFO,
       MCI_SYSINFO_QUANTITY | MCI_WAIT, &mmInfo, 0 );
      
      if ( rc )
      {
        CHECKED_FREE( outbuf );
        debugf( "Error querying number of MIDI devices (rc=%lu).\n", rc );
        debugf( "MIDI sequencer support is disabled.\n" );
        numMIDIouts = 0;
        midiOutNames = NULL;
      } else {
        numMIDIouts = atol( outbuf );
        if ( numMIDIouts )
        {
          CHECKED_CALLOC( sizeof( char * ) * numMIDIouts, midiOutNames );
        } else midiOutNames = NULL;
      }
      
      for ( i=1; i<=numWaveOuts; ++i )
      {
        waveOutNames[i-1] =
         getDeviceDescription( MCI_DEVTYPE_WAVEFORM_AUDIO, (USHORT)i );
         
        if ( waveOutNames[i-1] && *(waveOutNames[i-1]) )
        {
          debugf( "Waveform audio device #%lu: %s\n", i, waveOutNames[i-1] );
        }
      }
      
      for ( i=1; i<=numMIDIouts; ++i )
      {
        midiOutNames[i-1] =
         getDeviceDescription( MCI_DEVTYPE_SEQUENCER, (USHORT)i );
         
        if ( midiOutNames[i-1] && *(midiOutNames[i-1]) )
        {
          debugf( "MIDI sequencer device #%lu: %s\n", i, midiOutNames[i-1] );
        }
      }
      
      mmThreadsSize = 8;
      CHECKED_CALLOC( 8 * sizeof( ULONG ), mmThreads );
      CHECKED_CALLOC( 8 * sizeof( USHORT ), mmDeviceIDs );
      // Initially allow for 8 simultaneous multimedia threads
      
      DosCreateMutexSem( NULL, &mmMutex, 0, FALSE );
    }
  }
  
  debugf( "MultiMedia initialization is complete.\n" );
}

void deinitializeCompatibility( void )
{
  ULONG i;
  
  if ( numMmThreads )
  {
    ULONG rc, timePolled;
    
    if ( numMmThreads )
      debugf( "%d multimedia threads are still active.  Shutting them down.\n",
       numMmThreads );
    
    for ( i=0; i<numMmThreads; ++i )
    {
      if ( !mmThreads[i] ) continue;
      
      timePolled = 0;
      rc = DosWaitThread( mmThreads + i, DCWW_NOWAIT );
      while ( rc == ERROR_THREAD_NOT_TERMINATED && timePolled < 10 )
      {
        DosSleep( 1000 );
        if ( !mmThreads[i] ) break;
        timePolled++;
        rc = DosWaitThread( mmThreads + i, DCWW_NOWAIT );
      }
      // Give them 10 seconds to shut down nicely.
      
      if ( rc == ERROR_THREAD_NOT_TERMINATED && mmThreads[i] )
      {
        MCI_GENERIC_PARMS genParms;
        
        debugf( "Timed out waiting for thread to end.\n" );
        debugf( "Forcefully closing multimedia device ID %u.\n",
         mmDeviceIDs[i] );
         
        (*pMciSendCommand) ( mmDeviceIDs[i], MCI_CLOSE, MCI_WAIT,
         &genParms, 0 );
        
        debugf( "Killing thread %d.\n", mmThreads[i] );
        DosKillThread( mmThreads[i] );
      }
      // Then get nasty.
    }
  }
  
  if ( mmThreadsSize )
    CHECKED_FREE( mmThreads );
  
  if ( mmDeviceIDs )
    CHECKED_FREE( mmDeviceIDs );
  
  if ( mmMutex )
    DosCloseEventSem( mmMutex );
  
  if ( mmpmDLL )
    DosFreeModule( mmpmDLL );
  
  if ( mdmDLL )
    DosFreeModule( mdmDLL );
  
  pMciSendCommand = NULL;
  pMmioOpen = NULL;
  pMmioClose = NULL;
  mmpmDLL = 0;
  mdmDLL = 0;
  
  if ( numWaveOuts )
  {
    for ( i=0; i<numWaveOuts; ++i )
    {
      if ( waveOutNames[i] )
      {
        CHECKED_FREE( waveOutNames[i] );
        waveOutNames[i] = NULL;
      }
    }
    
    CHECKED_FREE( waveOutNames );
    waveOutNames = NULL;
    numWaveOuts = 0;
  }
  if ( numMIDIouts )
  {
    for ( i=0; i<numMIDIouts; ++i )
    {
      if ( midiOutNames[i] )
      {
        CHECKED_FREE( midiOutNames[i] );
        midiOutNames[i] = NULL;
      }
    }
    
    CHECKED_FREE( midiOutNames );
    midiOutNames = NULL;
    numMIDIouts = 0;
  }
  
  #ifdef MRM_DEBUG_MEMORY
  {
    int leaksFound = 0;
    
    debugf( "LEAK-CHECKING THROUGH %d BLOCKS.\n", numMemTracers );
    
    for ( i=0; i<numMemTracers; ++i )
    {
      if ( !memTracer[i].blockAddr ) continue;
      if ( !memTracer[i].freeModule )
      {
        leaksFound++;
        debugf( "MEMORY LEAK DETECTED!!\n" );
        debugf( "BLOCK 0x%08lx WAS ALLOCATED IN %s LINE %d.\n",
         memTracer[i].blockAddr, memTracer[i].allocModule,
         memTracer[i].allocLine );
      }
    }
    
    if ( leaksFound )
    {
      debugf( "FOUND %d MEMORY LEAKS.\n" );
    } else {
      debugf( "NO MEMORY LEAKS WERE FOUND.\n" );
    }
  }
  
  free( memTracer );
  numMemTracers = 0;
  allocMemTracers = 0;
  
  #endif
}

int isMultiMediaCapable( void )
{
  return mmpmDLL != NULLHANDLE;
}

void playThread( const char *fileName )
{
//  HMMIO mmioInst;
  MCI_OPEN_PARMS openParms = { 0 };
  MCI_PLAY_PARMS playParms = { 0 };
  MCI_GENERIC_PARMS genParms = { 0 };
  MMFORMATINFO formatInfo = { 0 };
  MCI_SET_PARMS setParms = { 0 };
  USHORT devType;
  FOURCC fcc;
  ULONG rc, mySlot;
  
  debugf( "Playing media file: %s\n", fileName );
  
  for ( mySlot=0; mySlot<mmThreadsSize; ++mySlot )
  {
    if ( mmThreads[ mySlot ] == *_threadid )
    {
      break;
    }
  }
  
  formatInfo.ulStructLen = sizeof( MMFORMATINFO );
  rc = (*pMmioIdentifyFile) ( (PSZ) fileName, NULL, &formatInfo, &fcc, 0, 0 );
  
  if ( rc )
  {
    debugf( "The multimedia subsystem had an error (%lu) identifying the media format for file: %s\n",
     rc, fileName );
    _endthread();
    return;
  }
  
  if ( formatInfo.ulMediaType == MMIO_MEDIATYPE_AUDIO )
    devType = MCI_DEVTYPE_WAVEFORM_AUDIO;
  else if ( formatInfo.ulMediaType == MMIO_MEDIATYPE_MIDI )
    devType = MCI_DEVTYPE_SEQUENCER;
  else {
    debugf( "Media type could not be played by Mr. Message: %s (type %lu)\n",
     fileName, formatInfo.ulMediaType );
    
    DosRequestMutexSem( mmMutex, SEM_INDEFINITE_WAIT );
    mmThreads[ mySlot ] = 0;
    mmDeviceIDs[ mySlot ] = 0;
    numMmThreads--;
    DosReleaseMutexSem( mmMutex );
    
    _endthread();
    return;
  }

/*
  mmioInst = (*pMmioOpen) ( fileName, NULL, MMIO_READ | MMIO_VERTBAR );
  
  if ( !mmioInst )
  {
    debugf( "The multimedia subsystem could not open or read this file: %s\n",
     fileName );
    
    DosRequestMutexSem( mmMutex, SEM_INDEFINITE_WAIT );
    mmThreads[ mySlot ] = 0;
    mmDeviceIDs[ mySlot ] = 0;
    numMmThreads--;
    DosReleaseMutexSem( mmMutex );
    
    _endthread();
    return;
  }
*/

  openParms.pszElementName = (PSZ) (fileName);
  openParms.pszDeviceType = (PSZ) (MAKEULONG( devType, 0 ));
  
  rc = (*pMciSendCommand) ( 0, MCI_OPEN, MCI_OPEN_TYPE_ID |
   MCI_OPEN_ELEMENT | MCI_OPEN_SHAREABLE | MCI_WAIT, &openParms, 0 );  
//   MCI_OPEN_MMIO | MCI_OPEN_SHAREABLE | MCI_WAIT, &openParms, 0 );
  
  mmDeviceIDs[ mySlot ] = openParms.usDeviceID;
  
  if ( rc )
  {
    debugf( "Error (%lu) opening the media control interface for file: %s\n",
     rc, fileName );
//    (*pMmioClose) ( mmioInst, 0 );

    DosRequestMutexSem( mmMutex, SEM_INDEFINITE_WAIT );
    mmThreads[ mySlot ] = 0;
    mmDeviceIDs[ mySlot ] = 0;
    numMmThreads--;
    DosReleaseMutexSem( mmMutex );
    
    _endthread();
    return;
  }
  
  setParms.ulAudio = MCI_SET_AUDIO_ALL;
  setParms.ulLevel = audioVolume;
  
  rc = (*pMciSendCommand) ( mmDeviceIDs[ mySlot ], MCI_SET,
   MCI_WAIT | MCI_SET_AUDIO | MCI_SET_VOLUME, &setParms, 0 );
  
  if ( !rc )
    rc = (*pMciSendCommand) ( openParms.usDeviceID, MCI_PLAY, MCI_WAIT,
     &playParms, 0 );
  
  if ( rc )
  {
    debugf( "The media control interface had an error (%lu) playing: %s\n",
     rc, fileName );
    (*pMciSendCommand) ( openParms.usDeviceID, MCI_CLOSE, MCI_WAIT,
     &genParms, 0 );
//    (*pMmioClose) ( mmioInst, 0 );
    
    DosRequestMutexSem( mmMutex, SEM_INDEFINITE_WAIT );
    mmThreads[ mySlot ] = 0;
    mmDeviceIDs[ mySlot ] = 0;
    numMmThreads--;
    DosReleaseMutexSem( mmMutex );
    
    _endthread();
    return;
  }
  
  (*pMciSendCommand) ( openParms.usDeviceID, MCI_CLOSE, MCI_WAIT,
   &genParms, 0 );
//  (*pMmioClose) ( mmioInst, 0 );
  
  debugf( "Done playing media file: %s\n", fileName );
  
  DosRequestMutexSem( mmMutex, SEM_INDEFINITE_WAIT );
  
  mmThreads[ mySlot ] = 0;
  mmDeviceIDs[ mySlot ] = 0;
    
  numMmThreads--;
  
  DosReleaseMutexSem( mmMutex );
}

void playMediaFile( const char *fileName )
{
  ULONG i;
  
  if ( !mmpmDLL ) return;
  
  DosRequestMutexSem( mmMutex, SEM_INDEFINITE_WAIT );
  
  numMmThreads++;
  if ( numMmThreads > mmThreadsSize )
  {
    i = mmThreadsSize;
    mmThreadsSize *= 2;
    CHECKED_REALLOC( mmThreads, sizeof( ULONG ) * mmThreadsSize, mmThreads );
    CHECKED_REALLOC( mmDeviceIDs, sizeof( USHORT ) * mmThreadsSize,
     mmDeviceIDs );
    
    for ( ; i<mmThreadsSize; ++i )
    {
      mmThreads[i] = 0;
      mmDeviceIDs[i] = 0;
    }
  }
  // Expand the number of simultaneous threads if needed.
  
  for ( i=0; i<mmThreadsSize; ++i )
  {
    if ( mmThreads[i] == 0 )
    {
      mmThreads[i] = _beginthread( (void (*) (void *))playThread, NULL,
       65536, (void *)fileName );
      break;
    }
  }
  // Find an empty slot and launch the thread.
  
  DosReleaseMutexSem( mmMutex );
}

void handleMrMessageEvent( UI_EventType eventType,
 sessionThreadSettings *specificSettings,
 sessionThreadSettings *globalSettings, int asynchronous )
{
  sessionThreadSettings *theSettings;
  
  if ( !specificSettings || specificSettings->inheritSettings )
    theSettings = globalSettings;
  else
    theSettings = specificSettings;
  
  if ( theSettings->settingFlags[ eventType ] & SET_SOUND_ENABLED )
  {
    playMediaFile( theSettings->sounds[ eventType ] );
  }
  
  // Handle REXX and shell script launching here too...
}

unsigned long testSettingsFlag( unsigned long theFlag,
 sessionThreadSettings *specificSettings,
 sessionThreadSettings *globalSettings )
{
  sessionThreadSettings *theSettings;
  
  if ( !specificSettings || specificSettings->inheritSettings )
    theSettings = globalSettings;
  else
    theSettings = specificSettings;
  
  return theSettings->sessionFlags & theFlag;
}

void setAudioVolume( ULONG volume )
{
  int i;
  
  if ( !mmpmDLL ) return;
  
  DosRequestMutexSem( mmMutex, SEM_INDEFINITE_WAIT );
  
  if ( volume > 100 ) audioVolume = 100;
  else audioVolume = volume;
  
  for ( i=0; i<mmThreadsSize; ++i )
  {
    if ( mmThreads[i] )
    {
      MCI_SET_PARMS setParms = { 0 };
      setParms.ulAudio = MCI_SET_AUDIO_ALL;
      setParms.ulLevel = audioVolume;
      (*pMciSendCommand) ( mmDeviceIDs[i], MCI_SET, MCI_WAIT | MCI_SET_AUDIO |
       MCI_SET_VOLUME, &setParms, 0 );
    }
  }
  // Set volume levels for all currently playing sounds/devices
  
  DosReleaseMutexSem( mmMutex );
}

ULONG getAudioVolume( void )
{
  return audioVolume;
}

